/*- Influence.cpp -------------------------------------------------------------*\
|
|   Contains:   All of the code for setting up vertex influencing in the Skeleton.
|
|   Written by: Tom Plunket, tomas@teleport.com:  TJP
|
|   Copyright:  (c) 1998 by Tom Plunket, all rights reserved
|
|   File history:
|        2/27/98    TJP
|       File created.
|
\*-----------------------------------------------------------------------------*/
#include "Globals.h"
#include <math.h>
#include <SimpObj.h>
#include <float.h>

// From \MaxSDK\Samples\GSphere\GSphere.cpp.
#define GSPHERE_CLASS_ID 0

//---------------------------
// DistanceToBox
//
// Gets the vertex distances from defOs to the boneOs boxobject or bounding box that is passed in.
static void DistanceToBox(ObjectState* defOs, Matrix3& nodeTM, ObjectState& boneOs, Matrix3& boneTM, double* distance)
{
    Matrix3 tm = Inverse(boneTM);//nodeTM / boneTM;
    Box3 box;
    Object* defObj = defOs->obj;
    int i;

    boneOs.obj->GetDeformBBox(0, box);

    for (i = 0 ; i < defObj->NumPoints() ; ++i)
    {
        Point3 pt = defObj->GetPoint(i);
        double dist = 0;

        pt = pt * tm;

        if (pt.x < box.pmin.x)
            dist += pow(box.pmin.x - pt.x, 2.0f);
        if (pt.x > box.pmax.x)
            dist += pow(pt.x - box.pmax.x, 2.0f);

        if (pt.y < box.pmin.y)
            dist += pow(box.pmin.y - pt.y, 2.0f);
        if (pt.y > box.pmax.y)
            dist += pow(pt.y - box.pmax.y, 2.0f);

        if (pt.z < box.pmin.z)
            dist += pow(box.pmin.z - pt.z, 2.0f);
        if (pt.z > box.pmax.z)
            dist += pow(pt.z - box.pmax.z, 2.0f);

        distance[i] = sqrt(dist);
    }
}

//---------------------------
// DistanceTo
//
// Gets the vertex distances from defOs to the boneOs that is passed in.
static void DistanceToSphere(ObjectState* defOs, Matrix3& nodeTM, ObjectState& boneOs, Matrix3& boneTM, double* distance)
{
//    Matrix3 tm = nodeTM;
    Object* defObj = defOs->obj;
    Interval interval = FOREVER;
    float radius;
    Point3 center;
    int i;

    // It just so happens that both GeoSphere and normal sphere
    // have as parameter block entry 0 the radius of the sphere.
    center = boneTM.GetTrans();
    ((SimpleObject*)boneOs.obj)->pblock->GetValue(0, 0, radius, interval);

    for (i = 0 ; i < defObj->NumPoints() ; ++i)
    {
        Point3 pt = defObj->GetPoint(i);// * tm;
        
        distance[i] = Length(pt - center) - radius;
        if (distance[i] < 0.0f)
            distance[i] == 0.0f;
    }
}

//---------------------------
// DistanceToCylinder
//
// Gets the vertex distances from defOs to the boneOs cylinder that is passed in.
// Cylinder is built around its local Z axis.
static void DistanceToCylinder(ObjectState* defOs, Matrix3& nodeTM, ObjectState& boneOs, Matrix3& boneTM, double* distance)
{
    Matrix3 tm = Inverse(boneTM);//nodeTM / boneTM;
    float radius, height;
    Object* defObj = defOs->obj;
    Interval iValid = FOREVER;
    int i;

    // Pblock entry 0 is the radius, and entry 1 is the height.
    ((SimpleObject*)boneOs.obj)->pblock->GetValue(0, 0, radius, iValid);
    ((SimpleObject*)boneOs.obj)->pblock->GetValue(0, 1, height, iValid);

    for (i = 0 ; i < defObj->NumPoints() ; ++i)
    {
        Point3 pt = defObj->GetPoint(i) * tm;
        double distFromZ = sqrt(pt.x*pt.x+pt.y*pt.y);
        double distFromRad = (distFromZ > radius) ? distFromZ - radius : 0.0f;

        if (pt.z < 0.0f)
        {
            // The point is "below" the cylinder
            if (distFromZ < radius)
                distance[i] = -pt.z;
            else
                distance[i] = sqrt(pt.z*pt.z + distFromRad*distFromRad);
        }
        else if (pt.z > height)
        {
            // the point is "above" the cylinder
            if (distFromZ < radius)
                distance[i] = pt.z - height;
            else
            {
                float dZ = pt.z - height;
                distance[i] = sqrt(dZ*dZ + distFromRad*distFromRad);
            }
        }
        else
            // the point is between the top and bottom of the cylinder.
            distance[i] = distFromRad;
    }
}

//---------------------------
// Influence
//
// Returns a float** that has all of the rational influences of each
// bone on each vertex.
double** Influence(INode* defNode, ObjectState* defOs, INodeTab& bones, float falloff)
{
    INode* bone;
    Matrix3 nodeTM, boneTM;
    Mesh mesh;
    ObjectState /*defOs,*/ boneOs;
    Object* obj, *defObj;
    double** distances;
    int i, j, boneCt, nPts;

    boneCt = bones.Count();
    distances = new double*[boneCt];
    if (distances == NULL)
    {
        TSTR errMsg;
        MessageBox(NULL, _T("Can't allocate space for the influencing data."), _T("Uh oh"), MB_OK);
        return NULL;
    }

//    defOs = node->EvalWorldState(0);
    nodeTM = defNode->GetObjectTM(0);
//    defObj = defOs.obj->ConvertToType(defObjClassID);
    defObj = defOs->obj;

    nPts = defObj->NumPoints();
    for (i = 0 ; i < boneCt ; ++i)
        distances[i] = new double[nPts];

    for (i = 0 ; i < boneCt ; ++i)
    {
        if (distances[i] == NULL)
        {
            while (i)
                delete [] distances[--i];

            delete [] distances;
            return NULL;
        }
    }

    // Loop through the bones and compute the distances from each vertex in
    // the object to that bone.
    for (i = 0 ; i < bones.Count() ; ++i)
    {
        bone = bones[i];
        boneOs = bone->EvalWorldState(0);
        boneTM = bone->GetObjectTM(0);
        obj = boneOs.obj;

#ifdef _DEBUG
        TCHAR* name = bone->GetName();
        DebugPrint(_T("Influencing from %s\n"), name);
#endif

        if (obj->SuperClassID() != GEOMOBJECT_CLASS_ID)
        {
            TSTR errMsg;
            errMsg.printf(_T("I don't know how to deal with objects like %s, so I'm removing it from the bones."), bone->GetName());
            MessageBox(NULL, errMsg, _T("Bone selection error"), MB_OK);

            bones.Delete(i, 1);
            --i;
            continue;
        }

        else if ( (obj->ClassID() == Class_ID(SPHERE_CLASS_ID, 0)) ||
                  (obj->ClassID() == Class_ID(GSPHERE_CLASS_ID, 0)) ||
                  (obj->ClassID() == Class_ID(GSPHERE_CLASS_ID, 32670)) )
            DistanceToSphere(defOs, nodeTM, boneOs, boneTM, distances[i]);
        else if (obj->ClassID() == Class_ID(CYLINDER_CLASS_ID, 0))
            DistanceToCylinder(defOs, nodeTM, boneOs, boneTM, distances[i]);
//        else if (obj->ClassID() == Class_ID(, 0))
        else
            DistanceToBox(defOs, nodeTM, boneOs, boneTM, distances[i]);
    }

    // Now invert all of the distances and "normalize" them
    // so that the sum of the inversions totals to one.  This
    // gives us a nice-and-pretty influence ratio from each
    // bone for each vertex.
    for (i = 0 ; i < nPts ; ++i)
    {
        double runTot = 0;
        const double maxVariance = 1.1;
        double lowest = DBL_MAX;
        double secondLowest = DBL_MAX;

        for (j = 0 ; j < boneCt ; ++j)
        {
            if (distances[j][i] < secondLowest)
            {
                if (distances[j][i] < lowest)
                {
                    secondLowest = lowest;
                    lowest = distances[j][i];
                }
                else
                    secondLowest = distances[j][i];
            }
        }

        for (j = 0 ; j < boneCt ; ++j)
        {
            if (distances[j][i] <= lowest)
                distances[j][i] = 1;
            else// if ((distances[j][i] / maxVariance) <= secondLowest)
                distances[j][i] = 1 / pow((distances[j][i]/lowest), falloff);
//            else
//                distances[j][i] = 0;
            runTot += distances[j][i];
        }

        for (j = 0 ; j < boneCt ; ++j)
            distances[j][i] = (float)(distances[j][i] / runTot);
    }

    return distances;
}
