/*- Skeleton.cpp --------------------------------------------------------------*\
|
|   Contains:   Definition for Modifier-derived class.
|
|   Written by: Tom Plunket, tomas@teleport.com:  TJP
|
|   Copyright:  (c) 1998 by Tom Plunket, all rights reserved
|
|   File history:
|        1/21/98    TJP
|       File created.
|
\*-----------------------------------------------------------------------------*/
#include "Skeleton.h"
#include "SkelRC.h"

extern HINSTANCE hInstance;

BOOL CALLBACK SkeletonDlgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);

//-------------------------------------
// GetBonesCallback
class GetBonesCallback : public HitByNameDlgCallback
{
public:
    GetBonesCallback() { nsn = NULL; }
    void SetNonShownNode(INode* node) { nsn = node; }
    void SetReturnNodes(INodeTab* nodes) { tab = nodes; }

    TCHAR *dialogTitle() { return _T("Select bones"); }
    TCHAR *buttonText() { return _T("These"); }
//    BOOL singleSelect() { return FALSE; }
//    BOOL useFilter() { return TRUE; }
    int filter(INode *node);
//    BOOL useProc() { return TRUE; }
    void proc(INodeTab &nodeTab);
//    BOOL doCustomHilite() { return FALSE; }
//    BOOL doHilite(INode *node) { return FALSE; }
//    BOOL showHiddenAndFrozen() { return FALSE; }

private:
    INodeTab* tab;
    INode* nsn;
};
static GetBonesCallback gbcb;

//-----------------
// HitByNameDlgCallback
// filter
int GetBonesCallback::filter(INode *node)
{
    return (nsn != node);
}

//-----------------
// HitByNameDlgCallback
// proc
void GetBonesCallback::proc(INodeTab &nodeTab)
{
    *tab = nodeTab;
}


//-------------------------------------
// Skeleton
//-----------------
// data
Interface* Skeleton::maxInterface = NULL;
HWND Skeleton::hPanel = NULL;
HWND Skeleton::comboBox = NULL;
ISpinnerControl* Skeleton::falloffSpinner = NULL;

#define OPERATION_INFLUENCE 0x1

//-----------------
// Skeleton
// constructor/destructor
Skeleton::Skeleton()
{
    pendingOperations = 0;
    influences = NULL;
    falloff = 2.0f;
}

Skeleton::~Skeleton()
{
    DeleteAllRefs();
}

//-------------
// Skeleton
// GetBones
void Skeleton::GetBones()
{
    INodeTab nodes;
    int i;

    GetNodes(this, &nodes);
    switch (nodes.Count())
    {
    default :
        MessageBox(NULL, _T("I don't know what to do when editing more than one node at a time."), _T("Help!"), MB_OK);

    case 1 :
        gbcb.SetNonShownNode(nodes[0]);
        break;

    case 0:
        MessageBox(NULL, _T("Hey, what's going on here?"), _T("Help!"), MB_OK);
        break;
    }

    if (influences)
    {
        for (i = 0 ; i < boneList.Count() ; ++i)
            delete [] influences[i];
        
        delete [] influences;
    }

    gbcb.SetReturnNodes(&nodes);
    if (!maxInterface->DoHitByNameDialog(&gbcb))
        return;

    DeleteAllRefsFromMe();
    boneList.ZeroCount();

    for (i = 0 ; i < nodes.Count() ; ++i)
        MakeRefByID(FOREVER, i, nodes[i]);

    boneList.Shrink();
    FillComboBox();
    pendingOperations |= OPERATION_INFLUENCE;

    NotifyDependents(FOREVER, GEOM_CHANNEL, REFMSG_CHANGE);
    maxInterface->RedrawViews(maxInterface->GetTime());
}

//-----------------
// Skeleton
// SpinnerChange
void Skeleton::SpinnerChange(short index)
{
/*
    switch (index)
    {
    case IDC_FALLOFF_SPINNER :
        falloff = falloffSpinner->GetFVal();
        pendingOperations |= OPERATION_INFLUENCE;
        break;
    }

    NotifyDependents(FOREVER, GEOM_CHANNEL, REFMSG_CHANGE);
    maxInterface->RedrawViews(maxInterface->GetTime());
*/
}

//-----------------
// Modifier
// NotifyInputChanged
void Skeleton::NotifyInputChanged(Interval changeInt, PartID partID, RefMessage message, ModContext *mc)
{
}

//-------------
// Modifier
// ModifyObject
void Skeleton::ModifyObject(TimeValue t, ModContext &mc, ObjectState* os, INode *node)
{
    Point3* newPoints;
//    Interval iValid = FOREVER;
    int numPoints, numBones;
    int i, j;
    Object* defObj;
    INode* bone;
    Matrix3 tm;

#ifdef _DEBUG
    TCHAR* name = node->GetName();
#endif

    if (pendingOperations & OPERATION_INFLUENCE)
    {
//        if (t == 0)
        {
            pendingOperations &= ~(OPERATION_INFLUENCE);
            influences = Influence(node, os, boneList, 10.0f);
        }
//        else
//            node->EvalWorldState(0);
    }

    // We don't need to proceed if we haven't got anything to influence or if we're at time 0.
    if ((influences != NULL) && (t != 0))// && (t != GetCOREInterface()->GetTime()))
    {
        //return;

        defObj = os->obj;
        numPoints = defObj->NumPoints();
        numBones = boneList.Count();
        newPoints = new Point3[numPoints];
        for (i = 0 ; i < numPoints ; ++i)
            newPoints[i].x = newPoints[i].y = newPoints[i].z = 0.0f;

        for (i = 0 ; i < numBones ; ++i)
        {
            bone = boneList[i];

    //        tm = node->GetObjectTM(0);
    //        tm.NoTrans();
    //        tm = tm * Inverse(bone->GetNodeTM(0)) * bone->GetNodeTM(t);
            tm = Inverse(bone->GetNodeTM(0)) * bone->GetNodeTM(t);

            for (j = 0 ; j < numPoints ; ++j)
            {
                Point3 pt = defObj->GetPoint(j) * tm;
                newPoints[j].x += (float)(pt.x * influences[i][j]);
                newPoints[j].y += (float)(pt.y * influences[i][j]);
                newPoints[j].z += (float)(pt.z * influences[i][j]);
            }
        }

        for (i = 0 ; i < numPoints ; ++i)
            defObj->SetPoint(i, newPoints[i]);

        delete [] newPoints;
    }

    os->obj->UpdateValidity(GEOM_CHAN_NUM, LocalValidity(t));
}

//-------------
// Modifier
// LocalValidity
Interval Skeleton::LocalValidity(TimeValue t)
{
    Interval valid = FOREVER;
    Matrix3 mat;
    int i;

    for (i = 0 ; i < boneList.Count() ; ++i)
    {
        boneList[i]->GetObjectTM(t, &valid);
    }

    return valid;
}

//-------------
// ReferenceTarget
// Clone
RefTargetHandle Skeleton::Clone(RemapDir &remap)
{
    return new Skeleton;
}

//-------------
// ReferenceMaker
// NotifyRefChanged
RefResult Skeleton::NotifyRefChanged(Interval changeInt, RefTargetHandle hTarget, PartID& partID, RefMessage message)
{
    switch (message)
    {
    case REFMSG_CHANGE:
        break;

    case REFMSG_TARGET_DELETED:
        boneList.Delete(FindRef(hTarget), 1);
    }

    //NotifyDependents(FOREVER, 0, REFMSG_CHANGE);
    return REF_SUCCEED; 
}

//-------------
// ReferenceMaker
// SetReference
void Skeleton::SetReference(int i, RefTargetHandle rtarg)
{
    if (i < boneList.Count())
        boneList[i] = (INode*)rtarg;
    else
        boneList.Append(1, (INode**)&rtarg, 10);
}

#define BONES_CHUNK 0x9378
#define CURRENT_VERSION 1
//-------------
// ReferenceMaker
// Save
IOResult Skeleton::Save(ISave *is)
{
    unsigned long nb;
    int val = CURRENT_VERSION;
    IOResult res;

    res = Modifier::Save(is);
    if (res != IO_OK)
        return res;

    is->BeginChunk(BONES_CHUNK);
    is->Write(&val, sizeof(int), &nb);
    val = boneList.Count();
    res = is->Write(&val, sizeof(int), &nb);
    is->EndChunk();

    return res;
}

//-------------
// ReferenceMaker
// Load
IOResult Skeleton::Load(ILoad *il)
{
    unsigned long nb;
    int nBones;
    IOResult res;
    int version;

    res = Modifier::Load(il);
    if (res != IO_OK)
        return res;

    while ((res = il->OpenChunk()) == IO_OK)
    {
        switch (il->CurChunkID())
        {
        case BONES_CHUNK :
            il->Read(&version, sizeof(int), &nb);
            il->Read(&nBones, sizeof(int), &nb);

            boneList.SetCount(nBones);
            while (nBones > 0)
                boneList[--nBones] = NULL;
            break;

        default :
            MessageBox(NULL, _T("Hey, Skeleton wants to read a chunk that doesn't belong to it!"), _T("Reading weirdness"), MB_OK);
        }

        res = il->CloseChunk();

        if (res != IO_OK)
            return res;
    }

    return IO_OK;
}

//-------------
// Animatable
// BeginEditParams
void Skeleton::BeginEditParams(IObjParam *ip, ULONG flags, Animatable *prev)
{
//    hPanel = ip->AddRollupPage(hInstance, MAKEINTRESOURCE(IDD_SKELETON_UI), BaseDlgProc, _T("Skeleton"), (LPARAM)this);
    hPanel = ip->AddRollupPage(hInstance, MAKEINTRESOURCE(IDD_SKELETON_UI), SkeletonDlgProc, _T("Skeleton"), (LPARAM)this);
    maxInterface = ip;
}

//-------------
// Animatable
// EndEditParams
void Skeleton::EndEditParams(IObjParam *ip, ULONG flags, Animatable *next)
{
    ip->DeleteRollupPage(hPanel);
    maxInterface = NULL;
}

//-----------------
// Skeleton
// FillComboBox
void Skeleton::FillComboBox()
{
    int i;
    SendMessage(comboBox, CB_RESETCONTENT, 0, 0);

    for (i = 0 ; i < boneList.Count() ; ++i)
    {
        TCHAR* name = boneList[i]->GetName();
        SendMessage(comboBox, CB_ADDSTRING, 0, (LPARAM)name);
    }

    SendMessage(comboBox, CB_INSERTSTRING, 0, (LPARAM)_T("< None >"));
    SendMessage(comboBox, CB_SETCURSEL, 0, 0);
}

//-----------------
// Skeleton
// InitRollup
void Skeleton::InitRollup(HWND hWnd)
{
    ICustButton* button;

    comboBox = GetDlgItem(hWnd, IDC_BONELIST);

    button = GetICustButton(GetDlgItem(hWnd, IDC_GET_BONES));
    button->SetType(CBT_CHECK);
    button->SetHighlightColor(GREEN_WASH);
    ReleaseICustButton(button);

    FillComboBox();

    falloffSpinner = SetupFloatSpinner(hWnd, IDC_FALLOFF_SPINNER, IDC_FALLOFF_EDIT, 0.01f, 100.0f, falloff);
}

//-----------------
// Skeleton
// DestroyRollup
void Skeleton::DestroyRollup()
{
    ReleaseISpinner(falloffSpinner);
}

//---------------------------
// SkeletonDlgProc
//BOOL Skeleton::DlgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
BOOL CALLBACK SkeletonDlgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static Skeleton* skel;
    ICustButton* button;

    switch (msg)
    {
    case WM_INITDIALOG :
        skel = (Skeleton*)lParam;
        skel->InitRollup(hWnd);
//        InitRollup(hWnd);
        break;

    case WM_DESTROY :
        skel = NULL;
        break;

    case WM_COMMAND :
        switch (LOWORD(wParam))
        {
        case IDC_GET_BONES :
            button = GetICustButton((HWND)lParam);
            skel->GetBones();
//            GetBones();
            button->SetCheck(FALSE);
            ReleaseICustButton(button);
            break;

        case IDC_BONELIST :
            if (HIWORD(wParam) == CBN_SELCHANGE)
;//                skel->ColorInfluencedVerts();
            break;
        }
        break;

    case CC_SPINNER_CHANGE :
//        SpinnerChange(LOWORD(wParam));
        skel->SpinnerChange(LOWORD(wParam));
        break;
    }

    return FALSE;
}
